@@ -601,6 +601,7 @@ QW[7]   USER/SW    USER/SW          |       |
  * The HFI1 packet header is consumed in many places and sometimes overloaded
  * for cache and memory allocation reasons.
  */
+
 union opx_hfi1_packet_hdr {
 	/* STL UNION */
 	union opx_hfi1_stl_packet_hdr {
@@ -926,15 +927,7 @@ union opx_hfi1_packet_hdr {
 
 	/*    OPX RMA RTS HEADER                   */
 	struct {
-		uint64_t reserved[2]; /* QW[0-1] */
-
-		/* QW[2] BTH (unused)*/
-		uint16_t reserved_1[3];
-		uint8_t	 origin_rx;
-		uint8_t	 reserved_2;
-
-		/* QW[3-4] BTH/KDETH */
-		uint64_t reserved_3[2];
+		uint64_t reserved[5]; /* QW[0-4] */
 
 		/* QW[5] SW */
 		uint8_t	 opcode;
@@ -1068,6 +1061,11 @@ union opx_hfi1_packet_hdr {
 	} __attribute__((__packed__)) service; /* "reliability service" */
 } __attribute__((__packed__)) __attribute__((__aligned__(8)));
 
+#define OPX_HDR_SLID_9B(_hdr) ((union opx_hfi1_packet_hdr *) _hdr)->lrh_9b.slid
+#define OPX_HDR_SLID_16B(_hdr)                                        \
+	(((union opx_hfi1_packet_hdr *) _hdr)->lrh_16b.slid20 << 20 | \
+	 ((union opx_hfi1_packet_hdr *) _hdr)->lrh_16b.slid)
+
 static_assert(sizeof(union opx_hfi1_packet_hdr) == sizeof(uint64_t[15]),
 	      "sizeof(union opx_hfi1_packet_hdr) must be 15 qwords!");
 
@@ -1135,11 +1133,14 @@ static inline size_t opx_hfi1_packet_hdr_payload_bytes(const union opx_hfi1_pack
 #define OPX_RC2_MASK 0b100
 #endif
 
-#define OPX_JKR_PRINT_16B_PBC(a)    opx_jkr_print_16B_pbc((a), __func__)
+#define OPX_PRINT_16B_PBC(a)	    opx_print_16B_pbc((a), __func__)
+#define OPX_PRINT_9B_PBC(a, b)	    opx_print_9B_pbc((a), b, __func__)
 #define OPX_JKR_PRINT_16B_LRH(a, b) opx_jkr_print_16B_lrh((a), (b), __func__)
 #define OPX_JKR_PRINT_16B_BTH(a, b) opx_jkr_print_16B_bth((a), (b), __func__)
 
-void opx_jkr_print_16B_pbc(uint64_t pbc1, const char *func);
+void opx_print_16B_pbc(uint64_t pbc1, const char *func);
+void opx_print_9B_pbc(uint64_t pbc1, const enum opx_hfi1_type hfi1_type, const char *func);
+
 void opx_jkr_print_16B_lrh(uint64_t lrh1, uint64_t lrh2, const char *func);
 void opx_jkr_print_16B_bth(uint64_t bth1, uint64_t bth2, const char *func);
 
@@ -1337,11 +1338,13 @@ static inline void fi_opx_hfi1_dump_packet_hdr(const union opx_hfi1_packet_hdr *
 
 #else
 // Disable the macros
-#define OPX_JKR_PRINT_16B_PBC(a)
+#define OPX_PRINT_16B_PBC(a)
+#define OPX_PRINT_9B_PBC(a, b)
 #define OPX_JKR_PRINT_16B_LRH(a, b)
 #define OPX_JKR_PRINT_16B_BTH(a, b)
 
-void opx_jkr_print_16B_pbc(uint64_t pbc1, const char *func);
+void opx_print_16B_pbc(uint64_t pbc1, const char *func);
+void opx_print_9B_pbc(uint64_t pbc1, const enum opx_hfi1_type hfi1_type, const char *func);
 void opx_jkr_print_16B_lrh(uint64_t lrh1, uint64_t lrh2, const char *func);
 void opx_jkr_print_16B_bth(uint64_t bth1, uint64_t bth2, const char *func);
 
@@ -1362,18 +1365,26 @@ static inline void fi_opx_hfi1_dump_packet_hdr(const union opx_hfi1_packet_hdr *
 		fi_opx_hfi1_dump_packet_hdr(__hdr, __hfi1_type, __func__, __LINE__); \
 	}
 
+#define OPX_DEBUG_PRINT_PBC(__pbc, __hfi1_type)        \
+	if (__hfi1_type & OPX_HFI1_CNX000) {           \
+		OPX_PRINT_16B_PBC(__pbc, __hfi1_type); \
+	} else {                                       \
+		OPX_PRINT_9B_PBC(__pbc, __hfi1_type);  \
+	}
+
 #define OPX_DEBUG_PRINT_PBC_HDR(__pbc, __hdr, __hfi1_type)                           \
 	if (__hfi1_type & OPX_HFI1_CNX000) {                                         \
-		OPX_JKR_PRINT_16B_PBC(__pbc);                                        \
+		OPX_PRINT_16B_PBC(__pbc);                                            \
 		OPX_JKR_PRINT_16B_LRH(__hdr->qw_16B[0], __hdr->qw_16B[1]);           \
 		OPX_JKR_PRINT_16B_BTH(__hdr->qw_16B[2], __hdr->qw_16B[3]);           \
 	} else {                                                                     \
+		OPX_PRINT_9B_PBC(__pbc, __hfi1_type);                                \
 		fi_opx_hfi1_dump_packet_hdr(__hdr, __hfi1_type, __func__, __LINE__); \
 	}
 
 #define OPX_DEBUG_PRINT_PBC_HDR_QW(q0, q1, q2, q3, q4, __hfi1_type) \
 	if (__hfi1_type & OPX_HFI1_CNX000) {                        \
-		OPX_JKR_PRINT_16B_PBC(q0);                          \
+		OPX_PRINT_16B_PBC(q0);                              \
 		OPX_JKR_PRINT_16B_LRH(q1, q2);                      \
 		OPX_JKR_PRINT_16B_BTH(q3, q4);                      \
 	}
@@ -1527,11 +1538,9 @@ union fi_opx_hfi1_packet_payload {
 		} noncontiguous;
 
 		struct {
-			uint64_t  ipc_handle[8];	     // Both AMD and NVIDIA handles are 64 bytes
-			uint64_t  src_device_id;	     // Needed on receiver for opening handle
-			uintptr_t origin_byte_counter_vaddr; // copy this value into cts packet that will be sent by
-							     // rank 1
-			enum fi_hmem_iface src_iface;	     // Needed on receiver for opening handle
+			uintptr_t	origin_byte_counter_vaddr; // copy into cts packet that will be sent
+			uint64_t	alignment_unused[2];	   // Align the IPC handle on a cacheline
+			struct ipc_info ipc_info;		   // Needed on receiver for opening handle
 		} ipc;
 	} rendezvous;
 
@@ -1561,24 +1570,26 @@ static_assert(sizeof(union fi_opx_hfi1_packet_payload) <= OPX_HFI1_MAX_PKT_SIZE,
 	      "sizeof(union fi_opx_hfi1_packet_payload) must be <= OPX_HFI1_MAX_PKT_SIZE!");
 static_assert(
 	offsetof(union fi_opx_hfi1_packet_payload, rendezvous.contiguous.immediate_byte) == FI_OPX_CACHE_LINE_SIZE,
-	"struct fi_opx_hfi1_packet_payload.rendezvous.contiguous.immediate_byte should be aligned on cacheline 1!");
+	"union fi_opx_hfi1_packet_payload.rendezvous.contiguous.immediate_byte should be aligned on cacheline 1!");
 static_assert(
 	offsetof(union fi_opx_hfi1_packet_payload, rendezvous.contiguous.immediate_block) == 2 * FI_OPX_CACHE_LINE_SIZE,
-	"struct fi_opx_hfi1_packet_payload.rendezvous.contiguous.immediate_block should be aligned on cacheline 2!");
+	"union fi_opx_hfi1_packet_payload.rendezvous.contiguous.immediate_block should be aligned on cacheline 2!");
 static_assert(offsetof(union fi_opx_hfi1_packet_payload, rendezvous.noncontiguous.iov) == 8,
-	      "struct fi_opx_hfi1_packet_payload.rendezvous.noncontiguous.iov should be 8 bytes into cacheline 0!");
+	      "union fi_opx_hfi1_packet_payload.rendezvous.noncontiguous.iov should be 8 bytes into cacheline 0!");
 static_assert(offsetof(union fi_opx_hfi1_packet_payload, rendezvous.noncontiguous.iov_ext) == FI_OPX_CACHE_LINE_SIZE,
-	      "struct fi_opx_hfi1_packet_payload.rendezvous.noncontiguous.iov_ext should be aligned on cacheline 1!");
+	      "union fi_opx_hfi1_packet_payload.rendezvous.noncontiguous.iov_ext should be aligned on cacheline 1!");
 static_assert(
 	OPX_HFI1_MAX_PKT_SIZE - sizeof(uintptr_t) - ((FI_OPX_MAX_HMEM_IOV) * sizeof(struct fi_opx_hmem_iov)) ==
 		FI_OPX_RZV_NONCONTIG_UNUSED,
-	"FI_OPX_RZV_NONCONTIG_UNUSED should be based on struct fi_opx_hfi1_packet_payload.rendezvous.noncontiguous!");
+	"FI_OPX_RZV_NONCONTIG_UNUSED should be based on union fi_opx_hfi1_packet_payload.rendezvous.noncontiguous!");
 #if FI_OPX_RZV_NONCONTIG_UNUSED
 static_assert(offsetof(union fi_opx_hfi1_packet_payload, rendezvous.noncontiguous.unused) +
 			      FI_OPX_RZV_NONCONTIG_UNUSED ==
 		      OPX_HFI1_MAX_PKT_SIZE,
-	      "struct fi_opx_hfi1_packet_payload.rendezvous.noncontiguous.unused should end at packet MTU!");
+	      "union fi_opx_hfi1_packet_payload.rendezvous.noncontiguous.unused should end at packet MTU!");
 #endif
+static_assert(offsetof(union fi_opx_hfi1_packet_payload, rendezvous.ipc.ipc_info.ipc_handle) == FI_OPX_CACHE_LINE_SIZE,
+	      "union fi_opx_hfi1_packet_payload.rendezvous.ipc.ipc_info.ipc_handle should be aligned on cacheline 1!");
 static_assert(
 	(offsetof(union fi_opx_hfi1_packet_payload, tid_cts.tidpairs) +
 	 sizeof(((union fi_opx_hfi1_packet_payload *) 0)->tid_cts.tidpairs)) == OPX_HFI1_MAX_PKT_SIZE,
@@ -1608,7 +1619,7 @@ struct fi_opx_hfi1_ue_packet {
 	uint64_t  tag;
 	opx_lid_t lid;
 	uint16_t  subctxt_rx;
-	uint8_t	  is_intranode;
+	uint8_t	  is_shm;
 
 	uint8_t unused;
 
