@@ -487,21 +487,31 @@ err:
 	return -errno;
 }
 
-struct fi_opx_global_data fi_opx_global = {.hfi_local_info.type	  = OPX_HFI1_UNDEF,
-					   .pkt_size		  = OPX_HFI1_DEFAULT_PKT_SIZE,
-					   .opx_hfi1_type_strings = {[OPX_HFI1_UNDEF]  = "OPX_HFI1_UNDEF",
-								     [OPX_HFI1_JKR_9B] = "CN5000-mixed",
-								     [OPX_HFI1_WFR]    = "OPA100",
-								     [3]	       = "ERROR",
-								     [OPX_HFI1_JKR]    = "CN5000",
-								     [5]	       = "ERROR",
-								     [6]	       = "ERROR",
-								     [7]	       = "ERROR",
-								     [OPX_HFI1_CYR]    = "CN6000",
-								     [9]	       = "ERROR",
-								     [10]	       = "ERROR",
-								     [11]	       = "ERROR",
-								     [OPX_HFI1_CNX000] = "CN5000|CN6000"}};
+struct fi_opx_global_data fi_opx_global = {.hfi_local_info.local_lids_size = -1U,
+					   .hfi_local_info.type		   = OPX_HFI1_UNDEF,
+					   .hfi_local_info.sim_rctxt_fd	   = -1,
+					   .hfi_local_info.sim_sctxt_fd	   = -1,
+					   .hfi_local_info.lid		   = (opx_lid_t) 0,
+					   .hfi_local_info.hfi_unit	   = (uint8_t) -1U,
+					   .hfi_local_info.sriov	   = false,
+					   .hfi_local_info.multi_vm	   = false,
+					   .hfi_local_info.multi_lid	   = false,
+					   .hfi_local_info.min_rctxt	   = 0,
+					   .hfi_local_info.max_rctxt	   = INT_MAX,
+					   .pkt_size			   = OPX_HFI1_DEFAULT_PKT_SIZE,
+					   .opx_hfi1_type_strings	   = {[OPX_HFI1_UNDEF]	= "OPX_HFI1_UNDEF",
+									      [OPX_HFI1_JKR_9B] = "CN5000-mixed",
+									      [OPX_HFI1_WFR]	= "OPA100",
+									      [3]		= "ERROR",
+									      [OPX_HFI1_JKR]	= "CN5000",
+									      [5]		= "ERROR",
+									      [6]		= "ERROR",
+									      [7]		= "ERROR",
+									      [OPX_HFI1_CYR]	= "CN6000",
+									      [9]		= "ERROR",
+									      [10]		= "ERROR",
+									      [11]		= "ERROR",
+									      [OPX_HFI1_CNX000] = "CN5000|CN6000"}};
 /* ROUTE CONTROL table for each packet type */
 int opx_route_control[OPX_HFI1_NUM_PACKET_TYPES];
 
@@ -791,6 +801,12 @@ static void do_static_assert_tests()
 
 OPX_INI
 {
+#if HAVE_OPX_DL
+	ofi_mem_init();
+	ofi_hmem_init();
+	ofi_monitors_init();
+#endif
+
 	fi_opx_count	       = 1;
 	fi_opx_global.progress = FI_PROGRESS_MANUAL;
 	fi_opx_set_default_info(); // TODO: fold into fi_opx_set_defaults
@@ -820,7 +836,7 @@ OPX_INI
 	fi_param_define(&fi_opx_provider, "reliability_service_usec_max", FI_PARAM_INT,
 			"The number of microseconds between pings for un-acknowledged packets. Defaults to 500 usec.");
 	fi_param_define(
-		&fi_opx_provider, "reliability_service_max_oustanding_bytes", FI_PARAM_INT,
+		&fi_opx_provider, "reliability_service_max_outstanding_bytes", FI_PARAM_INT,
 		"This setting controls the maximum number of bytes allowed to be in-flight (sent but un-ACK'd by receiver) per reliability flow (one-way communication between two endpoints). Valid values are in the range of 8192-150,994,944 (8KB-144MB), inclusive. Default setting is 7,340,032 (7MB).");
 	fi_param_define(
 		&fi_opx_provider, "reliability_max_uncongested_pings", FI_PARAM_INT,
@@ -960,8 +976,8 @@ OPX_INI
 
 	if (fi_log_enabled(fi_opx_global.prov, FI_LOG_TRACE, FI_LOG_FABRIC)) {
 		Dl_info dl_info;
-		if (dladdr((void *) fi_opx_ini,
-			   &dl_info)) { // Use the OPX_INI function as the symbol to get runtime info
+		if (dladdr((void *) fi_opx_check_info,
+			   &dl_info)) { // Use the fi_opx_check_info function as the symbol to get runtime info
 			FI_TRACE(fi_opx_global.prov, FI_LOG_FABRIC,
 				 "Using opx Provider: Library file location is %s.\n", dl_info.dli_fname);
 		} else {
