@@ -158,6 +158,7 @@ int ofi_get_src_addr(uint32_t addr_format,
 {
 	switch (addr_format) {
 	case FI_SOCKADDR:
+	case FI_SOCKADDR_IP:
 	case FI_SOCKADDR_IN:
 	case FI_SOCKADDR_IN6:
 		return fi_get_src_sockaddr(dest_addr, dest_addrlen,
@@ -216,6 +217,7 @@ int ofi_get_addr(uint32_t *addr_format, uint64_t flags,
 
 	switch (*addr_format) {
 	case FI_SOCKADDR:
+	case FI_SOCKADDR_IP:
 		sa_family = 0;
 		ret = fi_get_sockaddr(&sa_family, flags, node, service,
 				      (struct sockaddr **) addr, addrlen);
@@ -258,19 +260,7 @@ void *ofi_av_addr_context(struct util_av *av, fi_addr_t fi_addr)
 
 int ofi_verify_av_insert(struct util_av *av, uint64_t flags, void *context)
 {
-	if (av->flags & FI_EVENT) {
-		if (!av->eq) {
-			FI_WARN(av->prov, FI_LOG_AV, "no EQ bound to AV\n");
-			return -FI_ENOEQ;
-		}
-
-		if (flags & FI_SYNC_ERR) {
-			FI_WARN(av->prov, FI_LOG_AV, "invalid flag\n");
-			return -FI_EBADFLAGS;
-		}
-	}
-
-	if (flags & ~(FI_MORE | FI_SYNC_ERR)) {
+	if (flags & ~(FI_MORE | FI_SYNC_ERR | FI_FIREWALL_ADDR | FI_AV_USER_ID)) {
 		FI_WARN(av->prov, FI_LOG_AV, "unsupported flags\n");
 		return -FI_EBADFLAGS;
 	}
@@ -283,19 +273,45 @@ int ofi_verify_av_insert(struct util_av *av, uint64_t flags, void *context)
 	return 0;
 }
 
+int ofi_av_insert_addr_at(struct util_av *av, const void *addr, fi_addr_t fi_addr)
+{
+	struct util_av_entry *entry = NULL;
+
+	assert(ofi_genlock_held(&av->lock));
+	ofi_av_straddr_log(av, FI_LOG_INFO, "inserting addr", addr);
+	HASH_FIND(hh, av->hash, addr, av->addrlen, entry);
+	if (entry) {
+		if (fi_addr == ofi_buf_index(entry))
+			return FI_SUCCESS;
+
+		ofi_av_straddr_log(av, FI_LOG_WARN, "addr already in AV", addr);
+		return -FI_EALREADY;
+	}
+
+	entry = ofi_ibuf_alloc_at(av->av_entry_pool, fi_addr);
+	if (!entry)
+		return -FI_ENOMEM;
+
+	memcpy(entry->data, addr, av->addrlen);
+	ofi_atomic_initialize32(&entry->use_cnt, 1);
+	HASH_ADD(hh, av->hash, data, av->addrlen, entry);
+	FI_INFO(av->prov, FI_LOG_AV, "fi_addr: %" PRIu64 "\n",
+		ofi_buf_index(entry));
+	return 0;
+}
+
 int ofi_av_insert_addr(struct util_av *av, const void *addr, fi_addr_t *fi_addr)
 {
 	struct util_av_entry *entry = NULL;
 
-	assert(ofi_mutex_held(&av->lock));
-	ofi_straddr_log(av->prov, FI_LOG_INFO, FI_LOG_AV, "inserting addr", addr);
+	assert(ofi_genlock_held(&av->lock));
+	ofi_av_straddr_log(av, FI_LOG_INFO, "inserting addr", addr);
 	HASH_FIND(hh, av->hash, addr, av->addrlen, entry);
 	if (entry) {
 		if (fi_addr)
 			*fi_addr = ofi_buf_index(entry);
 		if (ofi_atomic_inc32(&entry->use_cnt) > 1) {
-			ofi_straddr_log(av->prov, FI_LOG_WARN, FI_LOG_AV,
-							"addr already in AV", addr);
+			ofi_av_straddr_log(av, FI_LOG_WARN, "addr already in AV", addr);
 		}
 	} else {
 		entry = ofi_ibuf_alloc(av->av_entry_pool);
@@ -320,7 +336,7 @@ int ofi_av_remove_addr(struct util_av *av, fi_addr_t fi_addr)
 {
 	struct util_av_entry *av_entry;
 
-	assert(ofi_mutex_held(&av->lock));
+	assert(ofi_genlock_held(&av->lock));
 	av_entry = ofi_bufpool_get_ibuf(av->av_entry_pool, fi_addr);
 	if (!av_entry)
 		return -FI_ENOENT;
@@ -345,9 +361,9 @@ fi_addr_t ofi_av_lookup_fi_addr_unsafe(struct util_av *av, const void *addr)
 fi_addr_t ofi_av_lookup_fi_addr(struct util_av *av, const void *addr)
 {
 	fi_addr_t fi_addr;
-	ofi_mutex_lock(&av->lock);
+	ofi_genlock_lock(&av->lock);
 	fi_addr = ofi_av_lookup_fi_addr_unsafe(av, addr);
-	ofi_mutex_unlock(&av->lock);
+	ofi_genlock_unlock(&av->lock);
 	return fi_addr;
 }
 
@@ -358,35 +374,6 @@ ofi_av_lookup_addr(struct util_av *av, fi_addr_t fi_addr, size_t *addrlen)
 	return ofi_av_get_addr(av, fi_addr);
 }
 
-int ofi_av_bind(struct fid *av_fid, struct fid *eq_fid, uint64_t flags)
-{
-	struct util_av *av;
-	struct util_eq *eq;
-
-	av = container_of(av_fid, struct util_av, av_fid.fid);
-	if (eq_fid->fclass != FI_CLASS_EQ) {
-		FI_WARN(av->prov, FI_LOG_AV, "invalid fid class\n");
-		return -FI_EINVAL;
-	}
-
-	if (!(av->flags & FI_EVENT)) {
-		FI_WARN(av->prov, FI_LOG_AV, "cannot bind EQ to an AV that was "
-			"configured for synchronous operation: FI_EVENT flag was"
-			" not specified in fi_av_attr when AV was opened\n");
-		return -FI_EINVAL;
-	}
-
-	if (flags) {
-		FI_WARN(av->prov, FI_LOG_AV, "invalid flags\n");
-		return -FI_EINVAL;
-	}
-
-	eq = container_of(eq_fid, struct util_eq, eq_fid.fid);
-	av->eq = eq;
-	ofi_atomic_inc32(&eq->ref);
-	return 0;
-}
-
 static void util_av_close(struct util_av *av)
 {
 	HASH_CLEAR(hh, av->hash);
@@ -400,13 +387,10 @@ int ofi_av_close_lightweight(struct util_av *av)
 		return -FI_EBUSY;
 	}
 
-	if (av->eq)
-		ofi_atomic_dec32(&av->eq->ref);
-
 	ofi_genlock_destroy(&av->ep_list_lock);
 
 	ofi_atomic_dec32(&av->domain->ref);
-	ofi_mutex_destroy(&av->lock);
+	ofi_genlock_destroy(&av->lock);
 
 	return 0;
 }
@@ -415,16 +399,16 @@ int ofi_av_close(struct util_av *av)
 {
 	int ret;
 
-	ofi_mutex_lock(&av->lock);
+	ofi_genlock_lock(&av->lock);
 	if (av->av_set) {
 		ret = fi_close(&av->av_set->av_set_fid.fid);
 		if (ret) {
-			ofi_mutex_unlock(&av->lock);
+			ofi_genlock_unlock(&av->lock);
 			return ret;
 		}
 		av->av_set = NULL;
 	}
-	ofi_mutex_unlock(&av->lock);
+	ofi_genlock_unlock(&av->lock);
 
 	ret = ofi_av_close_lightweight(av);
 	if (ret)
@@ -536,15 +520,14 @@ int ofi_av_init_lightweight(struct util_domain *domain, const struct fi_av_attr
 			    struct util_av *av, void *context)
 {
 	int ret;
-	enum ofi_lock_type ep_list_lock_type;
+	enum ofi_lock_type av_lock_type, ep_list_lock_type;
 
 	ret = util_verify_av_attr(domain, attr);
 	if (ret)
 		return ret;
 
-	av->prov = domain->prov;
 	ofi_atomic_initialize32(&av->ref, 0);
-	ofi_mutex_init(&av->lock);
+
 	av->av_fid.fid.fclass = FI_CLASS_AV;
 	/*
 	 * ops set by provider
@@ -553,10 +536,21 @@ int ofi_av_init_lightweight(struct util_domain *domain, const struct fi_av_attr
 	 */
 	av->context = context;
 	av->domain = domain;
+	av->prov = domain->prov;
 
+	av_lock_type = domain->threading == FI_THREAD_DOMAIN &&
+				       domain->control_progress ==
+					       FI_PROGRESS_CONTROL_UNIFIED ?
+			       OFI_LOCK_NOOP :
+			       OFI_LOCK_MUTEX;
+
+	ret = ofi_genlock_init(&av->lock, av_lock_type);
+	if (ret)
+		return ret;
+
+	ep_list_lock_type = ofi_progress_lock_type(domain->threading,
+						   domain->control_progress);
 
-	ep_list_lock_type = ofi_progress_lock_type(av->domain->threading,
-						   av->domain->control_progress);
 	ret = ofi_genlock_init(&av->ep_list_lock, ep_list_lock_type);
 	if (ret)
 		return ret;
@@ -580,34 +574,6 @@ int ofi_av_init(struct util_domain *domain, const struct fi_av_attr *attr,
 	return ret;
 }
 
-void ofi_av_write_event(struct util_av *av, uint64_t data,
-			int err, void *context)
-{
-	struct fi_eq_err_entry entry = { 0 };
-	size_t size;
-	ssize_t ret;
-	uint64_t flags;
-
-	entry.fid = &av->av_fid.fid;
-	entry.context = context;
-	entry.data = data;
-
-	if (err) {
-		FI_INFO(av->prov, FI_LOG_AV, "writing error entry to EQ\n");
-		entry.err = err;
-		size = sizeof(struct fi_eq_err_entry);
-		flags = UTIL_FLAG_ERROR;
-	} else {
-		FI_DBG(av->prov, FI_LOG_AV, "writing entry to EQ\n");
-		size = sizeof(struct fi_eq_entry);
-		flags = 0;
-	}
-
-	ret = fi_eq_write(&av->eq->eq_fid, FI_AV_COMPLETE, &entry,
-			  size, flags);
-	if ((size_t) ret != size)
-		FI_WARN(av->prov, FI_LOG_AV, "error writing to EQ\n");
-}
 
 /*************************************************************************
  *
@@ -626,9 +592,9 @@ static int ip_av_insert_addr(struct util_av *av, const void *addr,
 	int ret;
 
 	if (ofi_valid_dest_ipaddr(addr)) {
-		ofi_mutex_lock(&av->lock);
+		ofi_genlock_lock(&av->lock);
 		ret = ofi_av_insert_addr(av, addr, fi_addr);
-		ofi_mutex_unlock(&av->lock);
+		ofi_genlock_unlock(&av->lock);
 	} else {
 		ret = -FI_EADDRNOTAVAIL;
 		if (fi_addr)
@@ -677,20 +643,13 @@ int ofi_ip_av_insertv(struct util_av *av, const void *addr, size_t addrlen,
 					fi_addr ? &fi_addr[i] : NULL, context);
 		if (!ret)
 			success_cnt++;
-		else if (av->eq)
-			ofi_av_write_event(av, i, -ret, context);
 		else if (sync_err)
 			sync_err[i] = -ret;
 	}
 
 done:
 	FI_DBG(av->prov, FI_LOG_AV, "%d addresses successful\n", success_cnt);
-	if (av->eq) {
-		ofi_av_write_event(av, success_cnt, 0, context);
-		ret = 0;
-	} else {
-		ret = success_cnt;
-	}
+	ret = success_cnt;
 	return ret;
 }
 
@@ -709,9 +668,9 @@ int ofi_ip_av_insert(struct fid_av *av_fid, const void *addr, size_t count,
 				 count, fi_addr, flags, context);
 }
 
-static int ip_av_insertsvc(struct fid_av *av, const char *node,
-			   const char *service, fi_addr_t *fi_addr,
-			   uint64_t flags, void *context)
+int ofi_ip_av_insertsvc(struct fid_av *av, const char *node,
+			const char *service, fi_addr_t *fi_addr,
+			uint64_t flags, void *context)
 {
 	return fi_av_insertsym(av, node, 1, service, 1, fi_addr, flags, context);
 }
@@ -781,8 +740,8 @@ static int ip_av_nodesym_getaddr(struct util_av *av, const char *node,
 {
 	struct addrinfo hints, *ai;
 	void *addr_temp;
-	char name[FI_NAME_MAX];
-	char svc[FI_NAME_MAX];
+	char name[OFI_NAME_MAX];
+	char svc[OFI_NAME_MAX];
 	size_t name_len, n, s;
 	int ret, name_index, svc_index, count = (int)(nodecnt * svccnt);
 
@@ -819,7 +778,7 @@ static int ip_av_nodesym_getaddr(struct util_av *av, const char *node,
 	for (n = 0; n < nodecnt; n++) {
 		if (nodecnt == 1) {
 			strncpy(name, node, sizeof(name) - 1);
-			name[FI_NAME_MAX - 1] = '\0';
+			name[OFI_NAME_MAX - 1] = '\0';
 		} else {
 			snprintf(name + name_len, sizeof(name) - name_len - 1,
 				 "%zu", name_index + n);
@@ -828,7 +787,7 @@ static int ip_av_nodesym_getaddr(struct util_av *av, const char *node,
 		for (s = 0; s < svccnt; s++) {
 			if (svccnt == 1) {
 				strncpy(svc, service, sizeof(svc) - 1);
-				svc[FI_NAME_MAX - 1] = '\0';
+				svc[OFI_NAME_MAX - 1] = '\0';
 			} else {
 				snprintf(svc, sizeof(svc) - 1,
 					 "%zu", svc_index + s);
@@ -862,7 +821,7 @@ int ofi_ip_av_sym_getaddr(struct util_av *av, const char *node,
 	struct in_addr ip4;
 	int ret;
 
-	if (strlen(node) >= FI_NAME_MAX || strlen(service) >= FI_NAME_MAX) {
+	if (strlen(node) >= OFI_NAME_MAX || strlen(service) >= OFI_NAME_MAX) {
 		FI_WARN(av->prov, FI_LOG_AV,
 			"node or service name is too long\n");
 		return -FI_ENOSYS;
@@ -889,9 +848,9 @@ int ofi_ip_av_sym_getaddr(struct util_av *av, const char *node,
 				     svccnt, addr, addrlen);
 }
 
-static int ip_av_insertsym(struct fid_av *av_fid, const char *node,
-			   size_t nodecnt, const char *service, size_t svccnt,
-			   fi_addr_t *fi_addr, uint64_t flags, void *context)
+int ofi_ip_av_insertsym(struct fid_av *av_fid, const char *node,
+			size_t nodecnt, const char *service, size_t svccnt,
+			fi_addr_t *fi_addr, uint64_t flags, void *context)
 {
 	struct util_av *av;
 	void *addr;
@@ -934,9 +893,9 @@ int ofi_ip_av_remove(struct fid_av *av_fid, fi_addr_t *fi_addr,
 	 * Thus, we walk through the array backwards.
 	 */
 	for (i = count - 1; i >= 0; i--) {
-		ofi_mutex_lock(&av->lock);
+		ofi_genlock_lock(&av->lock);
 		ret = ofi_av_remove_addr(av, fi_addr[i]);
-		ofi_mutex_unlock(&av->lock);
+		ofi_genlock_unlock(&av->lock);
 		if (ret) {
 			FI_WARN(av->prov, FI_LOG_AV,
 				"removal of fi_addr %"PRIu64" failed\n",
@@ -946,6 +905,14 @@ int ofi_ip_av_remove(struct fid_av *av_fid, fi_addr_t *fi_addr,
 	return 0;
 }
 
+bool ofi_ip_av_is_valid(struct fid_av *av_fid, fi_addr_t fi_addr)
+{
+	struct util_av *av =
+		container_of(av_fid, struct util_av, av_fid);
+
+	return ofi_bufpool_ibuf_is_valid(av->av_entry_pool, fi_addr);
+}
+
 int ofi_ip_av_lookup(struct fid_av *av_fid, fi_addr_t fi_addr,
 		     void *addr, size_t *addrlen)
 {
@@ -969,8 +936,8 @@ ofi_ip_av_straddr(struct fid_av *av, const void *addr, char *buf, size_t *len)
 static struct fi_ops_av ip_av_ops = {
 	.size = sizeof(struct fi_ops_av),
 	.insert = ofi_ip_av_insert,
-	.insertsvc = ip_av_insertsvc,
-	.insertsym = ip_av_insertsym,
+	.insertsvc = ofi_ip_av_insertsvc,
+	.insertsym = ofi_ip_av_insertsym,
 	.remove = ofi_ip_av_remove,
 	.lookup = ofi_ip_av_lookup,
 	.straddr = ofi_ip_av_straddr,
@@ -992,7 +959,7 @@ static int ip_av_close(struct fid *av_fid)
 static struct fi_ops ip_av_fi_ops = {
 	.size = sizeof(struct fi_ops),
 	.close = ip_av_close,
-	.bind = ofi_av_bind,
+	.bind = fi_no_bind,
 	.control = fi_no_control,
 	.ops_open = fi_no_ops_open,
 };
